/**
   Program CheckVsCv ("Check against the Controlled Vocabulary")
   Version 0.37: start to add more buttons 20060930 - nothing done
   last edit 20060919 - basic phrase check seems to work - debugging output commented out
   @version 0.36 2006-09-21
   	  added check to allow known words with apostrophe-s (no case-insensitive check)
   @version 0.35 2006-09-15
   	  started modification to permit multi-word phrases to be treated as single words
   @version 0.3 2006-04-30
   @author Pat Cassidy
   A simple stand-alone window to check words in a text against a controlled vocabulary dictionary (with all allowed morphological forms)
      No corrections are suggested, and no interface with other programs is provided.

   Usage:
    (1) The text to be checked is pasted into the top window, after deleting all text already there.
    (2) The spelling dictionary is loaded, by clicking the "Use Dictionary File" button.  The default controlled vocabulary is file "baseCV.txt".
        As communities develop their own controlled vocabularies for definitions, this file will be supplemented and the commujnity file should be loaded at this point.
    (3) The check of the words in the proposed definition is performed by clicking the "Check Spelling".
          If any words are not recognized, a message is provided.  If the word is ignored,
          the next unrecognized word will be shown.  There is a provision to change the word
          shown, or to terminate the test to modify the text or the dictionary.
    (4) After the controlled vocabulary is modified, it can be reloaded (as in step 2) without restarting the program,
          and a check of words against the modified controlled vocabulary can be perfromed.
    (5)  See file "readmeCV.doc" for additional discussion of the program.

*/

import java.io.*;
import java.awt.*;
import javax.swing.*;
import java.awt.event.*;
import java.util.*;
import java.text.*;
import java.nio.channels.FileChannel;
import java.lang.Character;
import java.util.ArrayList;
import java.util.Collections;


public class CheckVsCV
{
   public static void main(String[] args)
   {
	  String defaultVocabulary = null;
	  if(args.length >= 1)
	       System.out.println("Argument passed: " + args[0] + ".");
      TextEditFrame frame = new TextEditFrame();
      frame.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
      frame.show();
   }
}

/**
   A frame with a text area and components for search/replace.
*/
class TextEditFrame extends JFrame
{
    boolean doTheBinsrch = true;
    boolean doListSearch = false;
    boolean searchProps = false;
    boolean ignoreFW = true; // flag to ignore case problem in first word
    boolean fwCase = false; // flag that there is a case problem in first word
    boolean wordIsSelected = false;  // flag that select text button was pressed since last spell check

    Date tday;
//    double diff;
//    double perMs;
//    long thirdLoopStart = 0;
    NumberFormat nf = NumberFormat.getInstance();

    int     stringIndex = -5;
//    String inFileName = "";
    String inFileTitle = "";
    String defaultFile = "baseCv.txt";
    int MAX_ERRS = 200;
    String[] errorMessage = new String[MAX_ERRS];
    int  errorCount = 0;
    int checkIterations = 0;
    ArrayList ssortCI;
    ArrayList pfwsortCI;
    int selLoc = 0;  // for finding text to change - first loc
                    // initialized at read - dictionary or change-text
    int selLen = 0;  // text to change - len

    JTextField dictInput;

// ***combo
    // initial version has a simple text input field (dictInput) with a default value
    //   to specify the Controlled vocabulary file name
    // To switch rapidly from baseCV.txt to SupplementedCV.txt to community CV,
    //   a pick list is better.
    //    Start to add pick list in a ComboBox 2006-05-13
// First version with ComboBox, use simple Combo Box. Not editable.
//   should be editable
    boolean useCombo = true; // if set false, only the default vocabulary file can be used since
                             //  the text field is not visible.
// *********************************************************************************
// *************   Add additional vocabulary file names here     *******************
// *********************************************************************************
//
    String[] cvs = {"baseCV.txt", "SupplementedCV.txt", "CommunityCV.txt"};
    int cvLen = cvs.length;
    private JComboBox cvCombo = new JComboBox();
// ***combo

 // To handle multi-word phrases,
 // the controlled vocabulary reader will return three array lists:
 //    all the non-blank lines (terms)
 //    the multi-word terms
 //    the individual words (possible duplicates) in the multi-word terms

    ArrayList[] dictWords = new ArrayList[3];
    ArrayList phraseWordsSorted;
    ArrayList[] pwIndices;

    int MAX_LINES = 100000;
    int bytesReadIn = 0;
    int lineCount = 0;
    int dictWordCount = 0;  // total number of non-blank lines (words or phrases) in the input vocabulary
    int phraseCount = 0;  // total number of phrases (lines with more than one word) in the input vocabulary
    long fileLength = -1;

   public TextEditFrame()
   {
      setTitle("CheckVsCV");
      setSize(DEFAULT_WIDTH, DEFAULT_HEIGHT);

      dictWords[0] = null;
      dictWords[1] = null;
      dictWords[2] = null;

      Container contentPane = getContentPane();
      Box cbox = Box.createVerticalBox();

      JPanel panel = new JPanel();
	  panel.setLayout(new GridLayout(1,4));

      JPanel actionPanel = new JPanel();
	  actionPanel.setLayout(new GridLayout(1,2));
      // add button, text fields and labels

      JButton replaceButton = new JButton("Replace");
      panel.add(replaceButton);
      replaceButton.setBackground(Color.yellow);
      replaceButton.addActionListener(new ReplaceAction());

      JButton checkButton = new JButton("Check Spelling");
      actionPanel.add(checkButton);
      checkButton.setBackground(Color.green);

      checkButton.addActionListener(new CheckSpellAction());

      JButton okButton = new JButton("Select Text to Replace");
      actionPanel.add(okButton);
      okButton.addActionListener(new IgnoreSpellAction());

// *** create center panel of text and buttons
      JPanel centerPanel = new JPanel();
	  centerPanel.setLayout(new GridLayout(3,1));

      // add text area with scroll bars to center panel

      textArea = new JTextArea(16, 60);
//   add test text to the input area
      textArea.setText("Select and delete this text and paste your definition here.\n");
      if(useCombo) textArea.append("Choose the controlled vocabulary from the menu.\n");
      textArea.append("Click the \"Use Dictionary File\" button to load the controlled vocabulary.\n");
      textArea.append("Then click the \"Check Spelling\" button.\n");

      JScrollPane scrollPane = new JScrollPane(textArea);

//      centerPanel.add(scrollPane);
      contentPane.add(scrollPane, BorderLayout.NORTH);


	   JPanel dictPanel = new JPanel( );
	   dictPanel.setLayout(new GridLayout(1,2));
	   dictPanel.setBackground(Color.pink);

      JButton dictButton = new JButton("     Use Dictionary File >>>");
   	  dictPanel.add(dictButton);
      dictButton.setBackground(Color.pink);
      dictButton.addActionListener(new GetDictAction());


// ***combo
//    cvCombo.setMaximumCount(20);
//    cvCombo.setEnabled(true);
      cvCombo.setEditable(true);

      for (int i = 0; i < cvLen; i++)
            cvCombo.addItem(cvs[i]);
      cvCombo.setSelectedItem("baseCV.txt");
/*
    cvCombo.addItem("baseCV.txt");
    cvCombo.addItem("SupplementedCV.txt");
    cvCombo.addItem("CommunityCV.txt");
*/

       dictInput = new JTextField("baseCV.txt", 40);
// for test 2006-05-13 use ComboBox
       if(useCombo) dictPanel.add(cvCombo);
       else dictPanel.add(dictInput);
// ***combo

      from = new JTextField("", 16);
      panel.add(from);

      JLabel toLabel = new JLabel("with", JLabel.CENTER);
      toLabel.setBackground(Color.red);
      panel.add(toLabel);

      to = new JTextField("", 16);
      panel.add(to);

      centerPanel.add(dictPanel);
      centerPanel.add(panel);
      centerPanel.add(actionPanel);
      contentPane.add(centerPanel, BorderLayout.CENTER);

// *** create report test panel
      reportArea = new JTextArea(20, 60);
      JScrollPane rscrollPane = new JScrollPane(reportArea);
      contentPane.add(rscrollPane, BorderLayout.SOUTH);


   }

   public static final int DEFAULT_WIDTH = 700;
   public static final int DEFAULT_HEIGHT = 700;

   private JTextArea textArea;
   private JTextArea reportArea;
   private JTextField from;
   private JTextField to;

   /**
      The action listener for the replace button.
   */
   private class ReplaceAction implements ActionListener
   {
      public void actionPerformed(ActionEvent event)
      {
        String f = from.getText();
        selLen = f.length();
        // first, provide for no prior selection
        if (!wordIsSelected)
          {
			selLoc = 0;
            selLoc = textArea.getText().indexOf(f);
            if(selLoc < 0)
              {
				  reportArea.append(" ***ERR*** Specified text not found!\n");
				  return;
			  }
          }
         if (selLoc >= 0 && f.length() > 0)
            textArea.replaceRange(to.getText(), selLoc,
               selLoc + selLen);
         selLoc = 0;  // reset
      }
   }


// Select button
// misleading name for action - it actualy selects text
// note selLoc is initialized outside of this action
   private class IgnoreSpellAction implements ActionListener
   {
      public void actionPerformed(ActionEvent event)
      {
         String f = from.getText();
         selLoc = textArea.getText().indexOf(f, selLoc + selLen);
         selLen = f.length();
         reportArea.append("** Select button pressed. Loc = " +
                       selLoc + "; len = " + selLen +".\n");
         if (selLoc >= 0)
          {
			wordIsSelected = true;
            textArea.setCaretPosition(selLoc);
            textArea.setSelectedTextColor(Color.blue);
            textArea.moveCaretPosition(selLoc + selLen);
            textArea.requestFocus();
            String seltext = textArea.getSelectedText();
            reportArea.append("** Selected text = " +
                        seltext + " at Loc = " + selLoc + ".\n");
	      }
      }
   }


   /**
      The action listener for the read-dictionary button.
   */
   private class GetDictAction implements ActionListener
   {
      public void actionPerformed(ActionEvent event)
      {
		 int uniquePWcount =0;  // number of unique words in phrases
		 selLoc = 0;  // initialize with new dictionary
		 boolean gotDict = false;
		 String dictin;
// here get Combo Box reading
// ***combo
// cvs[i]
/*  *** non-editable version
         int cvno = cvCombo.getSelectedIndex();
         if(useCombo) dictin = cvs[cvno];
*/
// Editable ComboBox
         if(useCombo) dictin = (String) cvCombo.getSelectedItem();
// *** end of use-combo ***
		 else dictin = dictInput.getText();
         dictin = dictin.trim();
         if (dictin.equals(""))
           {
             dictin = defaultFile;
             reportArea.append("Using default dictionary file " +
		                    dictin + ".\n");
			}
		 inFileTitle = dictin;  // for report
         try
           {
             dictWords = readInFile(dictin);
           }
         catch (IOException e)
           {
             e.printStackTrace();
             errorMessage[++errorCount] = "IO exception on opening dictionary file\n";
           }
         // should return null if an IO error
         if(dictWords[0] == null)
		   {
		     reportArea.append("Could not read dictionary file " +
		                    dictin + ".\n");
		   }
		 else gotDict = true;
		 if(gotDict)
		  {

// ***DEBUG*** list multi-word phrases 20060916

             if(dictWords[1] != null &&  dictWords[1].size() > 0)
             {
              pfwsortCI = new ArrayList(dictWords[1].size());
              for (int s = 0; s < dictWords[1].size(); s++)
               {
				// create StringCI array of first word of phrases to handle case where
				// phrase occurs at the beginning of a sentence
				String phrz = (String)dictWords[1].get(s);
				String[] phrzTokens = phrz.split(" ");
                StringCI pfw = new StringCI(phrzTokens[0]);  // first word of phrase into the array
		 	    pfwsortCI.add(pfw);

// ****************************************************************************
/*							// but first, if it's first word of phrase check capitalization
							if(ptndx == 0 && (isFirstWordAfterPeriod || (w ==0 && ignoreFW)))
							 {
							  // could be unrecognized because of capitalization or non-capitalization
							  if(phraseTokensNosp[0].compareToIgnoreCase(stokens[w]) == 0)   // same ignoring case?
							  {
								// difference is apparently due to case -- accept word
								isaCaseWord = true;
								isFirstWordAfterPeriod = true;  // may not be true, but functions as if true
								break;
							  }
							 }
*/
// **************************************************************************************
//  ***DEBUG***    echo dictionary for testing
//			    reportArea.append("phrase " + s + "> " + phrz + "\n");
		       }
			 }
/*
             if(dictWords[2] != null &&  dictWords[2].size() > 0)
             {
              for (int s = 0; s < dictWords[2].size(); s++)
               {
//     echo dictionary for testing
			    reportArea.append("phrase word " + s + "> " + (String)dictWords[2].get(s) + "\n");
			    int pwindex = ((Integer)dictWords[3].get(s)).intValue();
			    reportArea.append(" . . . occurs in phrase > " + (String)dictWords[1].get(pwindex) + "\n");
		       }
			 }
*/
// ***DEBUG*** list multi-word phrases

  // for search of multi-word phrases, now create a list of the words in the
  // multi-word phrases, with each word occurring only once.  Each occurrence must be referenced to
  //  all of the phrases in which it occurs
         phraseWordsSorted =  new ArrayList(dictWords[2].size());
         pwIndices = new ArrayList[dictWords[2].size()];
         for (int pw = 0; pw < dictWords[2].size(); pw++)
          {
			 String phraseWd = (String)dictWords[2].get(pw);
			 int pwndx = Collections.binarySearch(phraseWordsSorted, phraseWd);
			 if(pwndx < 0)   // missing?
			   {
              // failure to find word gives value -(insertion point) -1
		 		 int insrt = -(pwndx +1);
				 phraseWordsSorted.add(insrt, phraseWd);
			  }
		   }  // end of for loop to add this unique phrase word to a sorted list

          // at this point, the sorted list of phrase words is available,but the indices into
          // the phrases have not been added 20060916:2006
          uniquePWcount = phraseWordsSorted.size();   // get unique words as an int
          if(uniquePWcount > 0)  // actually, could be 1 only if there is one phrase with a repeated word, e.g. 'no no'
              {
                for (int upw = 0; upw < uniquePWcount; upw++)
                  {
					 String upWord = (String)phraseWordsSorted.get(upw);
					 pwIndices[upw] = new ArrayList();
					 // search for occurrences of this word in the full phrase word list
					 //   each word in the full list will point to the phrase it came from
					 //  in dictWords[1]
                    for (int pwi = 0; pwi < dictWords[2].size(); pwi++)
                     {
						String dw = (String)dictWords[2].get(pwi);
						// do nothing unless the word matches the unique word
						if(dw.compareTo(upWord) == 0)
						{
                          // add an entry to the phrase index for this unique word
                          // the dictWord[3] element of the same index is the index in
                          //  dictWords[1] for the phrase containing this word
                          pwIndices[upw].add((Integer)dictWords[3].get(pwi));
						}
                     }
                    // *** check: all unique words must have at least one index value
                    if(pwIndices[upw].size() < 1)
                     {
                       reportArea.append("***ERROR 400****  unique word [" + upWord +
		                    "] has no index value in the index array!!\n");

				     }
                  }
			  }
/*
// ****DEBUG**** report results of indexing unique words
                for (int upw = 0; upw < uniquePWcount; upw++)
                  {
					 String upWord = (String)phraseWordsSorted.get(upw);
                     reportArea.append("*** unique word [" + upWord +
		                    "] is in the phrases:\n");
		             for(int uwi = 0; uwi < pwIndices[upw].size(); uwi++)
		              {
                       reportArea.append("  >> " + dictWords[1].get(((Integer)pwIndices[upw].get(uwi)).intValue()) + "\n");
					  }
				   }

// ****DEBUG**** report results of indexing unique words
*/

            // create array for case-insensitive spell check
            {
              ssortCI = new ArrayList(dictWords[0].size());
              for (int s = 0; s < dictWords[0].size(); s++)
               {
                StringCI sci = new StringCI((String)dictWords[0].get(s));
		 	    ssortCI.add(sci);
//     echo dictionary for testing
//			    reportArea.append("s> " + (String)dictWords.get(s) + "\n");
		       }
		      Collections.sort(ssortCI);
		      // test sort
//		      for(int s = 0; s < ssortCI.size(); s++)
//		      {
//			   reportArea.append("sci> " + ((StringCI)ssortCI.get(s)).getString() + "\n");
//		      }
	        }

			reportArea.append("Dictionary " +
		              dictin + " was successfuly read in.\n");
            // reach here only if there is no error above
            System.out.println("\n*****\nFile " + dictin + " has " + fileLength + " bytes");
//            reportArea.append(">>> File " + dictin + " has " + fileLength + " bytes,\n");
            reportArea.append(">>> File " + dictin + " has " +
		               lineCount + " lines\n");
            reportArea.append("  . . . and " +
		                    + dictWordCount + " terms\n");
            reportArea.append("  . . . of which " +
		                    + phraseCount + " are multi-word terms\n");
		    boolean echoWords = false;
		    if(echoWords)
		    {
              reportArea.append("Term are:\n");
              // print out the terms?
              for(int w = 0; w < dictWordCount; w++)
                 reportArea.append(" " + (String)dictWords[0].get(w) + "\n");
            }
         } // end of if-got-dict (then report stats)
		 else  reportArea.append("Could not open dictionary file " +
		                    dictin + ".\n");
         reportArea.requestFocus();
      } // end of actionPerformed()
   }  // end of class GetDictAction (reads controlled vocabulary file)


/** CheckSpellAction performs the spelling check
   it uses the dictWords[] array - if null, it does nothing
      uses the textArea in the window for input of text to be checked
   it reads in the textArea  with text to be checked,
       breaks it into tokens,
       and checks each token against the dictionary, which is contained in the dictWords array.


   For multi-word phrases: 20060916
      if a word is not recognized (this could be due to case difference) then it is checked agains the
        list of unique words occurring in phrases
        if it is on that list, the phrases it opccurs in are retrieved
        that occurrence of the word is checked to see of it is in the same exact phrase
        of it is not in a phrase,
        Then if it is a case difference, that is reported


   before actually checking the spelling, this routine tests whether a word comes after the end
   of a sentence - so that capitalization is not considered an error.
   See **check** for actual check routine
*/
   private class CheckSpellAction implements ActionListener
   {

      public void actionPerformed(ActionEvent event)
      {
       // use spelling check file
       // highlight words not in dictionary
       // give option to ignore or change
       if(dictWords[0] == null)
        {
           reportArea.append(" *****ERROR***** No valid dictionary! *****\n");
           return;
		}
       boolean[] isEndWord;
	   checkIterations++;
	   // add period to input as end-token to pacify split function
       String inText = textArea.getText() + ".";
       int tlines =  textArea.getLineCount();
//       String[] stokens = inText.split("[^a-zA-Z_0-9]");
       String[] stokens = inText.split("[^a-zA-Z_0-9']");
       int tokenCount = stokens.length;
       // find token locations in the string
       int lastLoc = 0;
       int[] tindex = new int[tokenCount];
       isEndWord = new boolean[tokenCount];
       // get indices of the tokens
       for(int t=0; t < tokenCount; t++)
        {
		  int tlen = stokens[t].length();
		  if(tlen < 1) continue;
		  int tloc = inText.indexOf(stokens[t], lastLoc);
		  if (tloc < 0)   // error
		   {
             reportArea.append(" ***** INTERNAL ERROR ***** Token: " +
                  stokens[t] + " - not found in text!! *****\n");
             return;
		   }
		  tindex[t] = tloc;
		  lastLoc = tloc + tlen;
        }
// Check for period after word using tindex[].  IF found,
//  mark it as an 'end word' -- allow uppercase after end words
//  Could extend this to other sentence endings, but not yet
// Also, don't complain about a capitalized word after a colon
       for(int t=0; t < tokenCount; t++)
        {
		  isEndWord[t] = false;
		  int ttlen = stokens[t].length();
		  if(ttlen < 1) continue;
		  int tlc = tindex[t] + ttlen;
 // allow a close paren before a period in this test
		  if(inText.charAt(tlc) == ')' || inText.charAt(tlc) == ']')
		          tlc = tlc + 1;
		  if(inText.charAt(tlc) == '.' || inText.charAt(tlc) == ':')
		    {
               isEndWord[t] = true;
			}
// ***DEBUG***
//      if(isEndWord[t] == true)
//             reportArea.append(" >> " + stokens[t]  + " is an end word.\n");
        }
       ArrayList dList = null;
       ArrayList unkList = new ArrayList(); // holds unrecognized words
       ArrayList caseArray = new ArrayList();  // flags whether they are recognized
                                              // if case is ignored
//       int unrecognized = checkDef(inText, dList);
// Do the actual spell-check against the controlled vocabulary
// superfluous segment *** (for debug purposes??)
       int unknowns = 0;  // no significance; not used except to control this segment during debug
	   if(unknowns < 1)
	   {
         reportArea.append("\n\n -----\nChecking words against vocabulary " +
                 inFileTitle + ": iteration " +  checkIterations + ":\n");
         reportArea.append("Found " + tlines  + " line" +
                   (tlines > 1? "s":""));
         reportArea.append(" with " + tokenCount  + " tokens.\n");
         fwCase = false;
         if(tokenCount > 0)
           {
            // **check**
            for(int w = 0; w <tokenCount; w++)
             {
			   fwCase = false;
			  // to allow apostrophe, we must delete single-quote delimiters
			   if(isaNumber(stokens[w])) continue;
			   if(stokens[w].trim().length() < 1) continue;
			   if(stokens[w].substring(0, 1).compareTo("'") == 0)
			    {
			     stokens[w] = stokens[w].substring(1);
			     int newlen = stokens[w].length();
                 if(stokens[w].substring(newlen-1, newlen).compareTo("'") == 0)
                      stokens[w] = stokens[w].substring(0, newlen-1);
			    }
               // see if the word is in the dictionary
               if(dictWords[0] != null)
                {
                 int windex = Collections.binarySearch(dictWords[0], stokens[w]);
                 // windex < 0 signals absence from vocabulary
                 if(windex < 0)
                  {
//                     reportArea.append(" **> " + stokens[w] + " <** not in dictionary\n");
                    // do a case-insensitive search
                     StringCI stci = new StringCI(stokens[w]);
                     int scindex = Collections.binarySearch(ssortCI, stci);
                     // assume first word is capitalized and words after the end of a sentence
                     // are  also capitalized
					 Boolean isCase = new Boolean(false);
					 boolean isFirstWordAfterPeriod = false;
					 boolean isaCaseWord = false;
					 boolean hasApostropheS = false;
                     if(scindex >=0)  // case-insensitive search finds the word: this is a case difference
                      {
						 isCase = new Boolean(true);  // boolean object to add to an ArrayList
						 isaCaseWord = true;
//                       reportArea.append(" **> " + stokens[w] + " <** is a case problem\n");
						 if (w ==0 && ignoreFW)  // ignore upper case in first word
						           fwCase = true;
				      }
				     else  // not a case word, check for apostrophe-s;
				          //  no other apostrophe allowed unless it's in the dictionary
				     {
					   if(stokens[w].endsWith("'s"))
					    {
						   reportArea.append(" **> " + stokens[w] + " <** has ending apostrophe-s\n");
                           String baseWord = stokens[w].substring(0, stokens[w].length() - 2);
                           reportArea.append(" **> Base word is: " + baseWord + "\n");
                           windex = Collections.binarySearch(dictWords[0], baseWord);
                           if(windex >= 0) hasApostropheS = true; // no case-insensitive yet
                           else  // do case-insensitive check
                           {
                             // ***NOTE*** here we are *not* an apostrophe-s word
                             stci = new StringCI(baseWord);
                             scindex = Collections.binarySearch(ssortCI, stci);
                             if(scindex >=0)  // case-insensitive search finds the word: this is a case difference
                              {
                               isCase = new Boolean(true);  // boolean object to add to an ArrayList
                               isaCaseWord = true;
                               if (w ==0 && ignoreFW)  // ignore upper case in first word
						           fwCase = true;
	                          }
						   }
				        }  // end of has-an-apostrope-s
					 } // end else scindex < 0 (not a case word)
					// do other checks only if this is not an apostrophe-s word
                    if(!hasApostropheS)  // if an apostrophe-s word,ignore other checks to end of windex < 0
                     {

           // ***NOTE*** one detail- if a word is apostrophe-s but the base word is wrong case,
           // ***           we will get here for the check of first-word-after-period
//   alternative reason for upper case: first word of another sentence
			  // check for period after previous non-null token
			  // we check first word after period for all words, just in case it turns out
			  // to be the first word of a phrase
                     isFirstWordAfterPeriod = false;
				     for(int p = w -1; p >=0; p--)
					   {
						  if(stokens[p].length() < 1) continue;
						  // check only the first non-null word before incorrect case
						  if(isEndWord[p]) isFirstWordAfterPeriod = true;
 						  break;
					   }
// *******DEBUG 618********
                     if(isaCaseWord && !isFirstWordAfterPeriod && !fwCase)
						    reportArea.append("Case for word [" + stokens[w] + "] is incorrect\n");
// *******DEBUG 618********
					 // if is a case problem and not a first word, then it is a phrase word or a missing word
				     // **********************************************************
//
//
//  *********************************************************************************
//  *********************************************************************************
// *****************************  Check Phrases *********************************************************
//  ***********************  but only if the dictionary actually has phrases  **********************
				        boolean isaPhraseWord = false;  // default

				     // **** begin phrases check ***
             if(dictWords[1] != null &&  dictWords[1].size() > 0)
             {
				     // here test if this unrecognized word (stokens[w]) occurs in a phrase
						// isaPhraseWord  will be true only if the word in question is part of the full phrase
                        int pwd = Collections.binarySearch(phraseWordsSorted, stokens[w]);
                        boolean isaFirstWord = false;
                        if(pwd < 0 && (isFirstWordAfterPeriod || w == 0))  //  no match for case-sensitive search,
                        										// but could be capitalized first word
                        {
// *****DEBUG  CI search
//						    reportArea.append("Word [" + stokens[w] + "] is first word after period,not in phrase CS.\n");
// *****DEBUG  CI search
						  // see if it matches a first phrase word case-insensitive
						  StringCI pwci = new StringCI(stokens[w]);  // this token to a StringCI
                          // there may be more than one phrase with the same first word
                          ArrayList pfwUnsorted  = new ArrayList(pfwsortCI.size());
                          for(int pl = 0; pl < pfwsortCI.size(); pl++)
                          {
							 pfwUnsorted.add((StringCI)pfwsortCI.get(pl));
						  }
						  Collections.sort(pfwsortCI);
                          int pcindex = Collections.binarySearch(pfwsortCI, pwci);
                          int fwCount = 0;
                          int fwIndex[] = new int[pfwsortCI.size()];
                          if(pcindex >= 0)  // found it in a phrase
                           {
 							   // find out how many of the phrases have this word as
							   // first word (case-insensitive)
							   for(int pt = 0; pt < pfwUnsorted.size(); pt++)
							   {
							     StringCI pfw = (StringCI)pfwUnsorted.get(pt);
							     if(pwci.compareTo(pfw) == 0) fwIndex[fwCount++] = pt;
							   }
// *****DEBUG  CI search
                            if(fwCount < 1)
                            {
						    reportArea.append("****ERROR 684 **** [" + stokens[w] + "] is not found in phrase list scan.\n");
							}
                            String fwPhrase = (String)dictWords[1].get(fwIndex[0]);
						    reportArea.append("Word [" + stokens[w] + "] is first word of phrase [" +
						             fwPhrase + "], but incorrect case.\n");
// *****DEBUG  CI search
                          // need to check remainder of words in this phrase
                            String[] phraseTokens = fwPhrase.split(" ");
                            int stokNdx = 1;
                            isaFirstWord = true;
                           // a simpler comparison, makes no provision for extra spaces in the definition
                            for(int pt = 1; pt < phraseTokens.length; pt++)
                              {
								if(phraseTokens[pt].length() < 1) continue;
								if(phraseTokens[pt].compareTo(stokens[w + stokNdx++]) == 0) continue;
								isaFirstWord = false;
                              }
                            if(!isaFirstWord && fwCount > 1) // phrase did not match & there are others
                              {
							    // checking remainder of the sorted first word list, around the point where the word was found
// *****DEBUG  CI search
							   reportArea.append("Word [" + stokens[w] + " is NOT IN FIRST PHRASE.\n");
// *****DEBUG  CI search
							  }
  // *****NOTE this word may be first of more than one phrase!
  // *****   need to check other phrases - go down whole list of phrases to compare
                             // check other phrases
                             if(isaFirstWord) isaPhraseWord = true;
						   }  // end of pcindex >=0 , is in the first word list CI
					     } // end of pwd < 0 && (not recognized, but is first after a period
                        if(pwd >= 0)   // this word is in the phrase word list
                        {
/*
// ****DEBUG 631
                         reportArea.append("***NOTE 632*** unrecognized word [" +
                                       stokens[w] + "] being checked.\n");
// ****DEBUG 631
*/
                         // OK, now have to check to see if the rest of the phrase is there too
						// ***** temporary for test 20060916 ***************
						// it would be simpler to just look for a string match of the phrase with part of the sentence, but
						//  we want to permit the limited typographical variant of having more than one space between words
						//  in a phrase - a typographical error, but one that could be hard to find and correct if not
						//  explicitly recognized
						// ***NOTE - TODO *****this is still not a perfect test, since extra separators
						//   in the original full entry will
						//   not be distinguished from spaces, and words separated by such non-characters (such as ';')
						//   will appear contiguous in the phrase search -- needs to be refined
						// *** first get the phrase(s) of which this word is a part
						//   w is the index of the unrecognized word
						//
						int ptndx = -1;  // index of the unrecognized word in the given phrase
						// scan over all vocabulary phrases that have this word till we find one that matches
						//  the phrase the word is actually in
						// ********** TODO - not done 20060917:1206
                        for(int uwi = 0; uwi < pwIndices[pwd].size(); uwi++)
                         {
							String thisPhrase = (String)dictWords[1].get(((Integer)pwIndices[pwd].get(uwi)).intValue());
/*
// ****DEBUG 631
                         reportArea.append("***NOTE 657*** unrecognized word [" +
                                       stokens[w] + "] is in phrase [" + thisPhrase + "].\n");
// ****DEBUG 631
*/
                            String[] phraseTokens = thisPhrase.split(" ");
//                            String[] phraseTokensNosp = new String[phraseTokens.length];
                            String[] ptTempStr = new String[phraseTokens.length];
                            // find index of unrecognized word
                            // first normalize dictionary phrase to be sure no double spaces are in it
                            int ptni = 0;  // index used to add the normalized phrase words
// ****DEBUG  make string of the test tokens
                            String phrasePhrase = "";
                            for(int pt = 0; pt < phraseTokens.length; pt++)
                             {
                               if(phraseTokens[pt].length() < 1)
                                {
                                  reportArea.append("***WARNING 657*** there is an extra space in dictionary phrase [" +
                                       thisPhrase + "]\n");
									continue;
								}
                               ptTempStr[ptni++] = phraseTokens[pt];
                               phrasePhrase = phrasePhrase + phraseTokens[pt] + " ";
						     }
						     // copy tokens to final string array
						     String[] phraseTokensNosp = new String[ptni];
						     for(int pt = 0; pt < ptni; pt++)
						      {
						     	phraseTokensNosp[pt] = ptTempStr[pt];
						      }
/*
// ****DEBUG 714
                              reportArea.append("***NOTE 687 *** For unrecognized word [" +
                                       stokens[w] + "] testing normalized phrase string [" + phrasePhrase  + "].\n");
// ****DEBUG 714
*/
                            // here, phraseTokensNosp has the dictionary phrase without extra spaces
                            // now find the unrecognized phrase word
 						     ptndx = -1;  // index of the unrecognized word in the given phrase
                             for(int pt = 0; pt < phraseTokensNosp.length; pt++)
                              {
								if(phraseTokensNosp[pt].compareTo(stokens[w]) == 0)
								{
								  ptndx = pt;
								  break;
								}
						      }
						 	if(ptndx < 0)  // not found in this phrase?  error!
							 {
                              reportArea.append("***ERROR 703 *** phrase word [" + stokens[w] +
                                    "] was not found in phrase [" + thisPhrase + "]\n");
							  continue;
							 }
/*
// ****DEBUG 714
                         reportArea.append("***NOTE 699 *** For unrecognized word [" + stokens[w] +
                         "], index in normalized phrase string [" + phrasePhrase  + "] is " + ptndx + ".\n");
// ****DEBUG 714
*/
							// OK index found.  Now see if the unknown word is in the entry at the right point

							if(ptndx > w) continue;  // phrase word too close to beginning of definition to fit.
							int startndx = w - ptndx -3;  // allow up to 3 extra spaces in phrase
							if(startndx < 0) startndx = 0;
							// now normalize that section of the inpout definition to elimiate extra spaces
							// ***NOTE*** because of the simplified tokenization of the input definition,
							// other separators will show up as null tokens and cannot be distinguished from
							// spaces.  At this point 20060917, we only provide a warning.
							// ** one more precaution: see if last word is beyond lenght of the definiti0on
							int lastDefndx = startndx + phraseTokensNosp.length + 3;
							if (lastDefndx > tokenCount) lastDefndx = tokenCount; // actually limit would be token count, not last index
							String[] entryTokensNosp = new String[phraseTokensNosp.length + 3];
                            int etni = 0;  // index used to add the normalized words to definition segment
// ****DEBUG  make string of the test tokens
                            String testPhrase = "";
                            for(int pt = startndx; pt < lastDefndx; pt++)
                             {
                               if(stokens[pt].length() < 1)
                                {
                                  reportArea.append("***WARNING 697*** there is an extra space or symbol " +
                                    " in definition after tokens " +
                                       stokens[pt -2] + " " + stokens[pt -1] + "\n");
									continue;
								}
                               entryTokensNosp[etni++] = stokens[pt];
                               testPhrase = testPhrase + stokens[pt] + " ";
						     }
/*
// ****DEBUG 714
                         reportArea.append("***NOTE 714 *** For unrecognized word [" +
                                       stokens[w] + "] testing def string [" + testPhrase  + "].\n");
// ****DEBUG 714
*/
						    // the normalized phrases should match exactly, once the first word is found
							// now find first word of phrase, but don't pass maximum position
							// we are now using entryTokensNosp, rather than tokens array as the array to match
							int startWord = -1;
							for(int ti = 0; ti < 4; ti++)
							{
							  if(phraseTokensNosp[0].compareTo(entryTokensNosp[ti]) == 0)
							  {
								 startWord = ti;
								 break;
							  }
							}
                            if(startWord < 0) // first word not found, this cannot be the dictionary phrase
                                 continue;
                            // first word found; following words [index starts at 1 after the first word] must match the phrase exactly
                            // startWord is the index in the normalized entryTokensNosp array (segment of the definition)
                            boolean fullPhraseMatches = true;
/*
// ****DEBUG 752
                         reportArea.append("***NOTE 752 *** For unrecognized word [" +
                                       stokens[w] + "] testing def string [" + testPhrase  +
                                       "] with startWord [" + startWord + "].\n");
                         reportArea.append("***  and length of phrase tokens  [" + phraseTokensNosp.length + "].\n");
                            if(phraseTokensNosp == null)
                            {
                            reportArea.append("***ERROR 760*** null phraseTOkensNosp!.\n");
							}
                            else
                               reportArea.append("***NOTE 760*** phraseTOkensNosp.length =" + phraseTokensNosp.length + ".\n");
// ****DEBUG 752
*/
							for(int ti = 1; ti < phraseTokensNosp.length; ti++)
							{
							  if(phraseTokensNosp[ti].compareTo(entryTokensNosp[ti + startWord]) != 0)
							   {
								 fullPhraseMatches = false;
								 break;
							   }
							}
                            if (!fullPhraseMatches) continue;  // this phrase isn't the reason for the unknown word being there
                            isaPhraseWord = true;  // this phrase is in the definition in the right place, word is accepted
                            break;
                         }  // end of uwi scan loop over all phrases with this word
                         if(isaCaseWord && isFirstWordAfterPeriod) // both were set true if this is a case word at
                                                           // beginning of a legal phrase
								break;
                         // if word not found in a phrase, we fall through with isaPhraseWord still false
						} // end of pwd > 0, i.e. this word occurs in a multi-word phrase
                    // at this point, if the unrecognized word is part of a known phrase that is in the
                    // entry, isaPhraseWord will be true
				   } // end of checking phrase words (if phrases are in the dictionary)
//
//  *********************************************************************************
// *****************************  End of Check Phrases *********************************************************
// **********************************************************
//
//
                    if(!isaCaseWord && !isaPhraseWord)  // absolute unknown, not just a case variant isCase is false
					  {
                        unkList.add(stokens[w]);
                        unknowns++;
				        caseArray.add(isCase);   // keeps track of whether unkowns are absolute unknowns or just case variants;
				        				// the index will be the same as the index of the word in the unkList array
				      }
					 if(isaCaseWord && !isFirstWordAfterPeriod && !fwCase && !isaPhraseWord)
					  {
                        unkList.add(stokens[w]);
                        unknowns++;
				        caseArray.add(isCase);
				      }
				    } // end of if !hasApostropheS (i.e. this is a known word possessive
				  }  // end of if-windex < 0 (word is not in Controlled vocabulary)
		       } // end of if-dictwords not null (dictionary present)
		     } // end of w loop over tokens
		     if(unknowns < 1)
		      {
                reportArea.append("!! All words recognized.\n");
                from.setText("");
//  ***** focus after setting caret position
                String rt = reportArea.getText();
                int rtLen = rt.length();
                reportArea.setCaretPosition(rtLen);
                reportArea.requestFocus();
//  ***** focus after setting caret position
			  }
			 else  // loop over detected unknown words
			  {
				for(int w = 0; w< unknowns; w++)
				 {
				   // if the only error is a case problem in the first word, ignore it
				   //   and report no problems
				   if(w ==0 && fwCase == true)
				     {
						if(unknowns > 1) continue;
                        reportArea.append("!! All words recognized.\n");
                        from.setText("");
//  ***** focus after setting caret position
                String rt = reportArea.getText();
                int rtLen = rt.length();
                reportArea.setCaretPosition(rtLen);
                reportArea.requestFocus();
//  ***** focus after setting caret position
                        break;
					 }
				   String thisUnk = (String)unkList.get(w);
				   boolean caseOrNot = ((Boolean)caseArray.get(w)).booleanValue();
                   from.setText(thisUnk);
//                   this.show();
                   Object[] options = {
					   "Ignore Word",
					    "Modify text or Dictionary",
					   "List Unrecognized Words"};

                   String optMsg;
                   if(caseOrNot)
                      optMsg = "Word [" + thisUnk + "] is not the correct case. What next?";
                   else
                      optMsg = "Word [" + thisUnk + "] Not found in dictionary. What next?";
                   int selection = JOptionPane.showOptionDialog(null,
                     optMsg,
                     "Please choose an option",
                      JOptionPane.YES_NO_CANCEL_OPTION,
                      JOptionPane.QUESTION_MESSAGE,
                      null,
                      options,
                      options[0]);

                   // first option is ignore
                   if(selection == JOptionPane.NO_OPTION)
                   {
                      from.setText(thisUnk);
                      String f = from.getText();
                      selLoc = textArea.getText().indexOf(f);
                      selLen = f.length();
 //         reportArea.append("** Select button pressed. Loc = " +
 //                       selLoc + "; len = " + selLen +".\n");
                     if (selLoc >= 0)
                      {
                       textArea.setCaretPosition(selLoc);
                       textArea.setSelectedTextColor(Color.blue);
                       textArea.moveCaretPosition(selLoc + selLen);
                       textArea.requestFocus();
                      }
                     break;
				   }
                   else if(selection == JOptionPane.CANCEL_OPTION)
                   {
                     reportArea.append("\n******************************************\n");
                     reportArea.append("***** Listing " +
                           unknowns  + " Unrecognized Words  ***********\n");
                     reportArea.append("******************************************\n");
                     for(int u = 0; u< unknowns; u++)
		               {
                          reportArea.append((String)unkList.get(u));
                          if(((Boolean)caseArray.get(u)).booleanValue())
                                                  reportArea.append(" - (Case)");
                          reportArea.append("\n");
		               }
                     reportArea.append("***** End of Unrecognized Words  ***********\n\n");
//  ***** focus after setting caret position
                String rt = reportArea.getText();
                int rtLen = rt.length();
                reportArea.setCaretPosition(rtLen);
                reportArea.requestFocus();
//  ***** focus after setting caret position
                     break;
				   }
				   else continue;
			     }  // end of loop over unknown words - to select action
			  } // end of else- unknown > 1, therefore present options
			    //   to handle the unrecognozed words
	      } // end of if-tokens > 0

       } // end if-unknowns < 1 (dummy segment)
    }

/*
// this is a routine left over from some timing tests
   int checkDef(String inDef, ArrayList dictList)
    {
    	int unk = 0;
    	int strLoc;
    	int stringIndex;
    //   doing a binary search on the word list
    //   reportArea.append("** Doing binary search " + icount + " times:.\n");
        strLoc = Collections.binarySearch(dictList, inDef);
    	return strLoc;
    }
*/

  } // end class CheckSpellAction



     //Open a BufferedReader:
   static public BufferedReader open(String string)
   {
        FileInputStream f;
        try
        {
           f = new FileInputStream (string);
        }
        catch (FileNotFoundException e)
        {
           e.printStackTrace();
           System.exit(1);
           return null;
        }
        InputStreamReader inp = new InputStreamReader(f);
        BufferedReader b = new BufferedReader(inp);

        return b;
     }


     //Create a new PrintStream for output:
     static public PrintStream create(String string)
     {
        FileOutputStream f;
        try
        {
           f = new FileOutputStream(string);
        }
        catch (FileNotFoundException e)
        {
           e.printStackTrace();
           System.exit(1);
           return null;
        }
        PrintStream p = new PrintStream(f);
        return p;
     }

    // StringCI is just a string that has a compare function  which is
    // case-insensitive
    class StringCI implements Comparable
    {
	  String outstring;
  	 public StringCI(String instring)
	   {
         outstring = instring;
	   }

       public int compareTo(Object second)
       {
		  String firstIn = this.outstring;
//		  StringCI sec = new SringCI((String)second);
		  String secondIn = ((StringCI)second).getString();
		  return firstIn.compareToIgnoreCase(secondIn);
	   }
  	   public String getString()
  	   {
  	      return outstring;
  	   }
	}  // end of class StringCI
// note this is not StringsCI



// readInfile reads the words in a controlled vocabulary list,
//     one word (or phrase) per line, and puts them
//     into an ArrayList that is sorted alphabetically
//
    ArrayList[] readInFile(String inFileName)  throws IOException
     {
         FileInputStream f = null;

         FileChannel fc = null;
         BufferedReader inFile = null;
         boolean openOK = true;
         String inString; // the single line read in for each read call
         String trimString;  // inString after removal of leading and trailing whitespace
         int blockParentCount = 0;  // number of parents in a subclass block

         // reset global variables for repeat passes
         bytesReadIn = 0;   // to monitor progress of reading file
         lineCount = 0;  //  number of input lines, = number of lines returned in fileStrings
         String[] fileStrings = new String[MAX_LINES]; // the array of strings found in the file
         String[] phrases = new String[MAX_LINES]; // the array of strings in the file having more than one word
         dictWordCount = 0;   // total count of lines in the input vocabulary
         phraseCount = 0;   // total number of phrases in the input vocabulary
         int wc;   // for index of space in phrases

         try
         {
            f = new FileInputStream (inFileName);
            fc = f.getChannel();
         }
         catch (FileNotFoundException e)
         {
            e.printStackTrace();
            openOK = false;
         }
         // get size
         if(openOK && f != null)
         {
           try
           {
             fileLength = (int)fc.size();   // fileLength is used to set the sizes of the
                                            //  arrays for holding data from the input file
             InputStreamReader inp = new InputStreamReader(f);
             inFile = new BufferedReader(inp);       // global
  	       }
           catch (IOException e)
           {
             e.printStackTrace();
             errorMessage[++errorCount] = "IO exception on creating BufferedReader\n";
             return null;   // a negative number
           }
           // reach here only if there is no error above
         }  // end of If openOK && !null
         else   // report failure
         {
            System.out.println("\n*****\nFile " + inFileName + " could not be opened1!");
            errorMessage[++errorCount] = "IO exception on creating BufferedReader\n";
            return null;   // a negative number
  		              // don't ask again for the file name here:
  		              // the user can press the "read file" button again
  	     }
  	   // here we have successfully opened the file ("inFile").  Now read it by lines.
  	   //  readlines:   label -- perhaps we will use one for a labeled break??

         boolean lastLine = false;  // signal if the last line was non-null begin as true

// here we actually read in the words or phrases of the controlled vocabulary
// we use a reading loop to get the strings into a string array, and after
//  all words are read in, copy them to the output ArrayList
         while(true)    // use break to terminate loop
         {
  		   inString = inFile.readLine();  // first test, do nothing with the data
  		   if(inString == null)
  		      {
  			     break;      // use break instead of loop condition, as
  			                 //  we can't allow a null pointer to be used
  			                 //  in the String.equals() method below
  		      }
  		   lineCount++;
  		   if(lineCount >= MAX_LINES) break;
  		   else fileStrings[lineCount -1] = inString;
           bytesReadIn = bytesReadIn + inString.length();
           inString = inString.trim();
           if(inString.length() > 0)
             {
			   // permit and ignore comment lines starting with ';'
			   if(inString.charAt(0) != ';')
			    {
				 fileStrings[dictWordCount++] = inString;
				 if((wc = inString.indexOf(' ')) >= 0)
				   {
					   phrases[phraseCount++] = inString;
				   }
			    }
			 }

        }  // end of while loop over input lines
        ArrayList[] outStrings = new ArrayList[4];
        outStrings[0]= new ArrayList(dictWordCount); // the output with
             //              array size equal to number of dictionary words
        outStrings[1]= new ArrayList(phraseCount); // the list of multi-word phrases
        outStrings[2]= new ArrayList(3*phraseCount); // the list of individual words in the phrases
     // at this point, we only create the list of individual phrase words and for each, a pointer to the
     //  phrase that contains that word.  After these arrays are returned, if we want to
     //  sort this list (rather than search the words one by one), we will need to create
     //  arrays for each phrase word - each word may be in more than one phrase
        outStrings[3]= new ArrayList(3*phraseCount); // the list of indexes (integers) to the phrases for the
                                                     //   individual words in the vocabulary phrases

      for(int w = 0; w < dictWordCount; w++)
             outStrings[0].add(fileStrings[w]);
      int totalPhraseWords = 0;
      for(int w = 0; w < phraseCount; w++)
        {
			Integer pndx = new Integer(w); // to store it in an ArrayList
             outStrings[1].add(phrases[w]);
             ArrayList phraseWordList = getPhraseWords(phrases[w]);
             if(phraseWordList != null && phraseWordList.size() > 0)
			   {
				   for(int pww = 0; pww < phraseWordList.size(); pww++)
				   {
					   outStrings[2].add(phraseWordList.get(pww));
					   outStrings[3].add(pndx);           // index of the phrase for this phrase word
				   }
			   }
		}

// *** create index of lists of phrases for each phrase word
// 					   outStrings[3].add(phraseIndices

        // close input file
        try
         {
           inFile.close();
         }
        catch(IOException ie) {
           reportArea.append("** ERR 633 *** Can't close dictionary file !\n");
         }


     // Sort the input controlled vocabulary list.
     //  The input list should be sorted, but editing could cause loss of
     //   alphabetical order, so sort it here as a precaution
     //
        Collections.sort(outStrings[0]);
//        if(outStrings[2] != null) Collections.sort(outStrings[2]);

        return outStrings;
     }  // end of method readInFile()

// get the individual words in a phrase:
//  the only permitted separator is the space
    ArrayList getPhraseWords(String pString)
    {
	  ArrayList pw = new ArrayList();
	  if(pString == null)
	   {
           reportArea.append("** ERR 866 *** passed a null phrase line!\n");
           return null;
	   }
	  pString = pString.trim();
	  String[] stokens = pString.split(" ");
      if (stokens.length < 1) return null;
      for(int t = 0; t < stokens.length; t++)
      {
		// adjacent spaces will generate null strings using split() - watch out!
		  if(stokens[t].length() > 0)
                 pw.add(stokens[t]);
          else    // double space?
           {
             reportArea.append("*** WARNING 877 *** the vocabulary phrase:\n***  [" + pString + "]\n");
             reportArea.append("*** has a multi-space separator in it!\n\n");
		   }
	  }
	  return pw;
	}


// method to test if a string is a number.
// rather than use built-in methods, allow more
// flexibility here for future parsing extensions
    boolean isaNumber(String instr)
     {
		int len = instr.length();
		if (len < 1) return false;
		// check for valid characters.
		// anything invalid returns false
		for (int c = 0; c < len; c++)
		{
		  if(Character.isDigit(instr.charAt(c))) continue;
		  if(instr.charAt(c) == '.') continue;
		  if(instr.charAt(c) == ',') continue;
		  return false;
		}
		return true;
	 }

}  // end of class TextEditFrame
